import sys


class colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    RESET = '\033[0m'


def ok(msg, *a, **kw):
    print(f'{colors.OKGREEN}{msg}{colors.RESET}', *a, **kw)


def info(*args):
    print(*args)


def error(msg, *a, **kw):
    print(f'{colors.FAIL}{msg}{colors.RESET}', *a, file=sys.stderr, **kw)


def warning(msg, *a, **kw):
    print(f'{colors.WARNING}{msg}{colors.RESET}', *a, **kw)




import os
import pwd
from os import path, chown
from ipaddress import ip_address, IPv4Network

import yaml
try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader

from .linux import shell
from .console import ok, warning, info


USER_CONFIGURATIONFILE = '.ssh/tuntap.yml'


def yamlload(s):
    return yaml.load(s, Loader)


def gethost(name):
    user = pwd.getpwnam(name)
    configurationfilename = path.join(user.pw_dir, USER_CONFIGURATIONFILE)
    if not path.exists(configurationfilename):
        raise KeyError(name)

    with open(configurationfilename) as f:
        return yamlload(f), configurationfilename


def getallhosts():
    # Look for all linux users using pwd module
    # filter users which has home dirs
    for i in pwd.getpwall():
        if i.pw_uid >= 1000 and not i.pw_shell.endswith('nologin'):
            configurationfile = path.join(i.pw_dir, USER_CONFIGURATIONFILE)
            if not path.exists(configurationfile):
                continue

            with open(configurationfile) as f:
                conf = yamlload(f)

            yield i.pw_name, conf


def getallassignedaddresses():
    for u, c in getallhosts():
        yield ip_address(c['addresses']['client'])
        yield ip_address(c['addresses']['server'])


def assign(network):
    addresses = set(getallassignedaddresses())
    result = []

    # find two free addresses
    for ip in network.hosts():
        if ip in addresses:
            continue

        if len(result) >= 2:
            break

        result.append(ip)

    client, server = result
    return client, server, int(client) - int(network.network_address)


def getnetwork():
    return IPv4Network(settings.cidr)


def createinterface(host):
    index = str(host['index'])
    ifname = f'tun{index}'
    clientaddr = host['addresses']['client']
    serveraddr = host['addresses']['server']
    owner = host['remoteuser']
    lines = [f'{i}\n' for i in [
        f'allow-hotplug {ifname}',
        f'auto {ifname}',
        f'iface {ifname} inet static',
        f'  netmask 31',
        f'  pre-up ip tuntap add mode tun dev {ifname} user {owner} group {owner}',
        f'  pre-up ip address add dev {ifname} {serveraddr}/31 peer {clientaddr}/31',
    ]]

    ifacefilename = path.join('/etc/network/interfaces.d', ifname)
    with open(ifacefilename, 'w') as f:
        f.writelines(lines)

    ok(f'File {ifacefilename} has been created successfully.')

    shell(f'ip tuntap add mode tun dev {ifname} user {owner} group {owner}')
    shell(f'ip address add dev {ifname} {serveraddr}/31 peer {clientaddr}/31')
    shell(f'ip link set up dev {ifname}')


def addhost(network, user):
    configurationfile = path.join(user.pw_dir, USER_CONFIGURATIONFILE)
    if path.exists(configurationfile):
        warning(f'User is already exists: {user.pw_name}')
        warning(f'Overwriting the file: {configurationfile}')

    server, client, index = assign(network)
    info(f'Assigned addresses: {client} {server}')

    userconfiguration = dict(
        remoteuser=user.pw_name,
        uid=user.pw_uid,
        gid=user.pw_gid,
        shell=user.pw_shell,
        addresses=dict(client=str(client), server=str(server)),
        netmask=str(network.netmask),
        index=index
    )

    sshdirectory = path.dirname(configurationfile)
    if not path.exists(sshdirectory):
        os.mkdir(sshdirectory, mode=0o700)
        chown(sshdirectory, user.pw_uid, user.pw_gid)

    with open(configurationfile, 'w') as f:
        yaml.dump(userconfiguration, f, default_flow_style=False)

    chown(configurationfile, user.pw_uid, user.pw_gid)
    createinterface(userconfiguration)


def deletehost(username):
    host, configurationfilename = gethost(username)
    index = str(host['index'])
    ifname = f'tun{index}'
    shell(f'ifdown {ifname}')

    # remove the interfaces.d file
    ifacefilename = path.join('/etc/network/interfaces.d', ifname)
    if path.exists(ifacefilename):
        os.remove(ifacefilename)

    shell(f'ip tuntap delete {ifname} mode tun')

    # remove the .ssh/tuntap.yml file
    if path.exists(configurationfilename):
        os.remove(configurationfilename)




import os
import pwd
from os import path
from ipaddress import IPv4Network

import pymlconf
from easycli import SubCommand, Argument, Root

from .console import info, ok, error, warning
from . import network
from . import linux


DEFAULT_CIDR = '192.168.22.0/24'
DEFAULT_CONFIGURATIONFILENAME = f'/etc/sshtuntap.yml'
BUILTIN_CONFIGURATION = f'''
cidr: {DEFAULT_CIDR}
'''

settings = pymlconf.DeferredRoot()


def configure():
    settings.initialize(BUILTIN_CONFIGURATION, context=os.environ)


class InfoCommand(SubCommand):
    __command__ = 'info'

    def __call__(self, args):
        print(f'Configuration file: {args.configurationfilename}')
        print(f'CIDR: {settings.cidr}')


class SetupCommand(SubCommand):
    __command__ = 'setup'
    __arguments__ = [
        Argument(
            'cidr',
            nargs='?',
            default=DEFAULT_CIDR,
            help=f'The network/mask (aka CIDR), default: {DEFAULT_CIDR}'
        )
    ]

    def __call__(self, args):

        settings.cidr = args.cidr
        with open(args.configurationfilename, 'w') as f:
            f.write(settings.dumps())

        ok(f'Settings are saved into {args.configurationfilename}')


class HostAddCommand(SubCommand):
    __command__ = 'add'
    __aliases__ = ['a']
    __arguments__ = [
        Argument('name'),
        Argument(
            '-m', '--mode',
            choices=['tun', 'tap'],
            default='tun',
            help='default: tun'
        )
    ]

    def __call__(self, args):
        hostname = args.name
        try:
            host = pwd.getpwnam(hostname)
        except KeyError:
            error(f'Host {hostname} is not exists, please create it first.')
            return 1

        network.addhost(IPv4Network(settings.cidr), host)
        ok(f'Host {hostname} was created successfully')


class HostListCommand(SubCommand):
    __command__ = 'list'
    __aliases__ = ['l']

    def __call__(self, args):
        for u, c in network.getallhosts():
            addrs = c['addresses']
            info(addrs['client'], addrs['server'], u)


class HostDeleteCommand(SubCommand):
    __command__ = 'delete'
    __aliases__ = ['d', 'del']
    __arguments__ = [
        Argument('name'),
    ]

    def __call__(self, args):
        hostname = args.name
        try:
            network.deletehost(hostname)
        except KeyError:
            error(f'Host: {hostname} is not exists.')
        else:
            ok(f'Host {hostname} has been deleted successfully')


class ServerRoot(Root):
    __aliases__ = ['s']
    __completion__ = True
    __arguments__ = [
        Argument(
            '-c', '--configurationfilename',
            metavar='FILENAME',
            default=DEFAULT_CONFIGURATIONFILENAME,
            help=f'default: {DEFAULT_CONFIGURATIONFILENAME}'
        ),
        Argument('-V', '--version', action='store_true'),
        InfoCommand,
        SetupCommand,
        HostAddCommand,
        HostListCommand,
        HostDeleteCommand,
    ]

    def _execute_subcommand(self, args):
        filename = args.configurationfilename
        configure()

        if path.exists(filename):
            settings.loadfile(filename)

        elif args.command not in ('setup', 'install'):
            error(f'Configuration file does not exists: {filename}')
            return 1

        super()._execute_subcommand(args)

    def __call__(self, args):
        if args.version:
            from sshtuntap import __version__ as version
            print(version)
            return

        super().__call__(args)


def main():
    ServerRoot().main()




import os
import time
import socket
import struct
from os import path
import argparse
import traceback
from subprocess import CalledProcessError

import pymlconf
from easycli import SubCommand, Argument, Root

from .console import info, ok, error, warning
from .linux import shell



HOME = os.environ['HOME']
USER = os.environ['USER']
DEFAULT_CONFIGURATIONFILENAME = f'{HOME}/.ssh/tuntap.yml'
BUILTIN_CONFIGURATION = f'''
hostname:
localuser:
'''

settings = pymlconf.DeferredRoot()


class SSHRetryError(Exception):
    pass

def configure():
    settings.initialize(BUILTIN_CONFIGURATION, context=os.environ)


class InfoCommand(SubCommand):
    __command__ = 'info'

    def __call__(self, args):
        print(f'Configuration file: {args.configurationfilename}')
        print(f'CIDR: {settings.cidr}')


class SetupCommand(SubCommand):
    __command__ = 'setup'
    __arguments__ = [
        Argument('hostname', help=f'The server\'s hostname')
    ]

    def __call__(self, args):
        if USER == 'root':
            error('Do not run this command by the root user')
            return 1

        filename = args.configurationfilename
        hostname = args.hostname
        shell(f'scp {hostname}:.ssh/tuntap.yml {filename}')
        settings.loadfile(filename)
        settings.hostname = \
            hostname.split('@')[1] if '@' in hostname else hostname
        settings.localuser = USER

        with open(filename, 'w') as f:
            f.write(settings.dumps())

        ok(f'Settings are saved into {filename}')


class ConnectCommand(SubCommand):
    __command__ = 'connect'
    __aliases__ = ['c']
    __arguments__ = [
        Argument('-v', '--verbose', action='store_true', help='Verbose'),
        Argument(
            '-i', '--aliveinterval',
            type=int,
            default=2,
            help=\
                'ssh ServerAliveInterval option, see man ssh_config. ' \
                'default is: 2'
        ),
        Argument(
            '-m', '--alivecountmax',
            type=int,
            default=1,
            help=\
                'ssh ServerAliveCountMax option, see man ssh_config. ' \
                'default is: 1'
        ),
        Argument(
            '-r', '--retrymax',
            type=int,
            default=0,
            help='Maximum retry, default is: 0, infinite!'
        ),
        Argument(
            '--connecttimeout',
            type=int,
            default=2,
            help='ssh ConnectTimeout option, default is: 2.'
        ),
        Argument(
            '--connectionattempts',
            type=int,
            default=2,
            help='ssh ConnectionAttempts option, default is: 2.'
        )
    ]

    def getdefaultgateway(self):
        """Returns the current default gateway from `/proc`
        """
        with open('/proc/net/route') as fh:
            for line in fh:
                fields = line.strip().split()
                if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                    continue
                return socket.inet_ntoa(struct.pack("<L", int(fields[2], 16)))

    def connect(self, gateway, hostaddr, hostname, args):
        index = settings.index
        hostname = settings.hostname
        remoteuser = settings.remoteuser
        localuser = settings.localuser
        serveraddr = settings.addresses.server
        c = args.retrymax
        infinite = c == 0
        sshargs = []

        if args.verbose:
            sshargs.append('-v')

        sshargs.append(f'-o"ServerAliveInterval {args.aliveinterval}"')
        sshargs.append(f'-o"ServerAliveCountMax {args.alivecountmax}"')
        sshargs.append(f'-o"ConnectTimeout {args.connecttimeout}"')
        sshargs.append(f'-o"ConnectionAttempts {args.connectionattempts}"')

        while infinite or (c > 0):
            try:
                shell(f'ip route replace {hostaddr} via {gateway}')
                shell(f'ip route replace default via {serveraddr}')
                shell(
                    f'sudo -u {localuser} ssh {remoteuser}@{hostname} ' \
                    f'-Nw {index}:{index} {" ".join(sshargs)}'
                )
            except CalledProcessError as ex:
                if ex.returncode < 0:
                    break

            except:
                traceback.print_exc()
                time.sleep(3)

            finally:
                shell(f'ip route del {hostaddr} via {gateway}', check=False)
                shell(f'ip route replace default via {gateway}', check=False)

            time.sleep(1)
            c -= 1

    def __call__(self, args):
        if USER != 'root':
            error('Please run this command as root.')
            return 1

        hostname = settings.hostname
        remoteuser = settings.remoteuser
        localuser = settings.localuser
        index = settings.index
        ifname = f'tun{index}'
        clientaddr = settings.addresses.client
        serveraddr = settings.addresses.server
        netmask = settings.netmask
        hostaddr = socket.gethostbyname(hostname)

        gateway = self.getdefaultgateway()
        if gateway is None:
            error(f'Invalid default gateway: {gateway}')

        try:
            shell(f'ip tuntap add mode tun dev {ifname} user {localuser} group {localuser}')
            shell(
                f'ip address add dev {ifname} {clientaddr}/{netmask} ' \
                f'peer {serveraddr}/{netmask}'
            )
            shell(f'ip link set up dev {ifname}')
            self.connect(gateway, hostaddr, hostname, args)
        finally:
            info('Terminating...')
            shell(f'ip tuntap delete mode tun dev {ifname}', check=False)


class ClientRoot(Root):
    __aliases__ = ['c']
    __completion__ = True
    __arguments__ = [
        Argument(
            '-c', '--configurationfilename',
            metavar='FILENAME',
            default=DEFAULT_CONFIGURATIONFILENAME,
            help=f'default: {DEFAULT_CONFIGURATIONFILENAME}'
        ),
        Argument('-V', '--version', action='store_true'),
        SetupCommand,
        ConnectCommand,
    ]

    def _execute_subcommand(self, args):
        filename = args.configurationfilename
        configure()

        if path.exists(filename):
            settings.loadfile(filename)

        elif args.command not in ('setup', ):
            error(f'Configuration file does not exists: {filename}')
            return 1

        super()._execute_subcommand(args)

    def __call__(self, args):
        if args.version:
            from sshtuntap import __version__ as version
            print(version)
            return

        super().__call__(args)


def main():
    ClientRoot().main()

